# `combinators-p`

Program with promises in a functional style.

## Synopsis

    import {getJson, storeDb} from './async-utils';
    import {flow as flowP} from 'combinators-p';

    const url = "https://url.horse/api";

    const apiCall = flow([getJson, storeDb]);

    await apiCall(url);

## Get started

    npm install --save combinators-p

## Contents

<!-- START doctoc generated TOC please keep comment here to allow auto update -->
<!-- DON'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->


- [API](#api)
  - [isPromise](#ispromise)
  - [future](#future)
  - [constant](#constant)
  - [delay](#delay)
  - [fmap](#fmap)
  - [apply](#apply)
  - [caught](#caught)
  - [all](#all)
  - [tap](#tap)
  - [flatmap](#flatmap)
  - [lift2](#lift2)
  - [spread](#spread)
  - [flow](#flow)
  - [fold](#fold)
  - [whenElse](#whenelse)
  - [flatmap2](#flatmap2)
  - [compose](#compose)
  - [flatmap3](#flatmap3)
  - [unlessElse](#unlesselse)
  - [flatmap4](#flatmap4)
  - [retry](#retry)
  - [flatmap5](#flatmap5)
  - [when](#when)
  - [unless](#unless)
  - [retry2](#retry2)
  - [retry3](#retry3)
  - [map](#map)
  - [retry4](#retry4)
  - [map2](#map2)
  - [map3](#map3)
  - [map4](#map4)
  - [map5](#map5)

<!-- END doctoc generated TOC please keep comment here to allow auto update -->

## API

<!-- Generated by documentation.js. Update this documentation by updating the source code. -->

### isPromise

Test if an object is a promise.

**Parameters**

-   `p` **any** The object to test.

**Examples**

```javascript
const p = future(23);
isPromise(p); // Returns true;
```

Returns **[boolean](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Boolean)** Returns `true` if the object is a promise, otherwise
`false`;

### future

Lift a value into a promise. This is equivalent to `Promise.resolve`.

**Parameters**

-   `x` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** The value to lift into a promise. This can
    either be a value, or a promise that resolves to a value.

**Examples**

```javascript
const x = future(23);
const f = a => future(a + 1);
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** The value inside a promise.

### constant

Create a function that always returns the same value.

**Parameters**

-   `x` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** The value to return. This can either be a value
    or a promise for a value.

**Examples**

```javascript
const f = constant("Hello");
f().then(console.log); // Returns "Hello"
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** A promise that resolves to `x`.

### delay

Delay the resolution of a promise chain.

**Parameters**

-   `ms` **[number](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Number)** The time to wait in milliseconds.
-   `value` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** The promise or value to resolve after the
    delay.

**Examples**

```javascript
delay(100).then(console.log) // Waits 100 ms.
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)** A promise that resolved to value, or whatever value
resolves to.

### fmap

Map a function over a promise.

**Parameters**

-   `f` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)> | [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)&lt;any>)** The function to apply. `f` can
    either be a function, or a promise that resolves to a promise. The function
    can either return a value, or a promise that resolves to a promise.
-   `x` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** The value to apply to `f`. This can either be a
    value, or a promise that resolves to a value.

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** The result of `x` applied to `f`.

### apply

Apply a function wrapped in a promise to a promisified value.

**Parameters**

-   `pf` **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)>** A promise that resolves to a function.
-   `p` **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;Any>** A promise that resolves to a value.

**Examples**

```javascript
const pf = future(v => v + 1);
const p = future(1);
apply(pf, f).then(console.log); // Returns 2.
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;Any>** A promise resolving to x applied to the function
that f resolves to.

### caught

Catch an exception on a promise and call a exception handler. This is
equivalent to `Promise.catch`.

**Parameters**

-   `f` **[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)** The handler to call when the promise `p` get's
    rejected.
-   `p` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** A value or a promise that get's resolved.

**Examples**

```javascript
const f = () => new Error("Boom");
caught(console.error, f()); // Prints the exception.
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** A promise that resolves either to the value of `p`
or to the return value of `f`.

### all

Create a function that evaluates all promises in an array when called. This
is equivalent to `Promise.all`, with the difference that it creates a
callable function.

**Parameters**

-   `xs` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)> | [Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)&lt;any>)** An array of values that are
    resolved. `xs` can either be an array, or a promise that resolves to an
    array. Each element of `xs` can either be a value, or a promise that
    resolves to a value.

**Examples**

```javascript
const f = all([openFile1(), opeFile2(), openFile3()]);
f().then(console.log); // Returns [a, b, c];
```

Returns **[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)&lt;Promise.Array&lt;any>>** A function, that when called, resolves all
promises and returns an Array of values.

### tap

Run a function for side effect and return the original value. The original
value gets cloned and therefore can be modified in the tap handler.

**Parameters**

-   `f` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)> | [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function))** The function for side effect. The
    return value of this function will be ignored.
-   `x` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** The value to use for side effect. This can
    either be a value or a promise that resolves to a value.

**Examples**

```javascript
const f = a => future(a);
flow([f, tap(console.log)])(23); // Print "23" to the console.
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** Returns `x`.

### flatmap

Map a function over every element of a list and concatenate the results
into a single list. Only one promise at a time get's resolved.

**Parameters**

-   `f` **[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)&lt;any>** The function that is applied to every element. This
    function can either return a value or the promise for a value.
-   `xs` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)> | [Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)&lt;any>)** The list to map `f` over. This can
    either be an array, or the promise for an array.

**Examples**

```javascript
const f = x => [x, x];
const xs = [1, 2];
flatmap(f, xs).then(console.log); // Prints [1, 1, 2, 2];
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)>** The concatenation of applying every element of
`xs` to `f`.

### lift2

Lift a binary function over two promises.

**Parameters**

-   `f` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)> | [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)&lt;any, any>)** A binary function. This
    can either be a function, or a promise that resolves to a function. The
    function can either return a value, or a promise that resolves to a value.
-   `x` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** A value that gets lifted as the first argument
    of `f`. This can either be a value, or a promise that resolves to a value.
-   `y` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** A value that gets lifted as the first argument
    of `f`. This can either be a value, or a promise that resolves to a value.

**Examples**

```javascript
const f = (x, y) => future(x = y);
const a = future(1);
const b = future(2);
lift2(f, a, b).then(console.log); // Returns 3.
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** The value that f returns when applied to `x` and
`y`.

### spread

Call a variadic function with the value of a promise as it's arguments. If
the value is an array, flatten it to the formal parameters of the
fulfillment handler.

**Parameters**

-   `f` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)> | [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)&lt;any>)** The function to apply to the
    value of p. This function can either return a value or the promise for a
    value. This function can also be a promise that resolves to a function.
-   `p` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** The promise that resolves to the arguments for
    the function. This can either be a single value, an Array or a promise
    resolving to any of those.

**Examples**

```javascript
const add = (x, y) => x + y;
const p = future([1, 2]);
spread(add, p).then(console.log); // Prints 3
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** The result of applying the value of `p` to `f`.

### flow

Create a function out of a list of functions, where each successive
invocation is supplied the return value of the previous function call. The
new function forms a pipe where the results flow from left to right so to
speak. It's a shortcut for composing more than two functions.

**Parameters**

-   `fs` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)> | [Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)>)** An array of functions to
    compose. `fs` can either be an array, or a promise that resolves to an
    array. Each function can either return a value or a promise for a value.
-   `x` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise) | any)** The argument to call the function pipeline with. It
    can either be any value, or the promise for any value.

**Examples**

```javascript
const f = (x, y) => future(x + y);
const fs = map(f, [...Array(5).keys()]);
flow(fs, 0).then(console.log); // The sum of [0, 1, 2, 3, 4], returns 10
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** The result of applying `x` to the pipeline of
`fs`.

### fold

Reduce a list of values to a single value, using a reduction function. This
is equivalent to `Array.reduce`.

**Parameters**

-   `f` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)> | [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)&lt;any>)** The reduce function to map
    over the list. `f` can either be a function, or a promise that resolves to
    a promise. This function can either return a value or the promise for a
    value.
-   `acc` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any> | any)** The initial value to apply to the first call
    of `f`. It can either be any value or the promise for any value.
-   `xs` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)> | [Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)&lt;any>)** The list to reduce to a single
    value. This can either be an array, or the promise for an array.

**Examples**

```javascript
const f = (acc, x) => future(acc + x);
const xs = [...Array(5).keys()];
fold(f, 0, xs).then(console.log); // The sum of [0, 1, 2, 3, 4], returns 10;
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** The value of `xs` reduced over `f`.

### whenElse

Make a conditional test and either call the left or the right branch.

**Parameters**

-   `predicate` **[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)** Test if a condition is `true` or `false`.
-   `consequent` **[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)** Apply `value` to this function if the
    predicate return `true`;
-   `alternative` **[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)** Apply `value` to this function if the
    predicate return `false`;
-   `value` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise) | any)** Test this value to decide whether to call the
    consequent or the alternative.

**Examples**

```javascript
const pred = userExists;
const cons = updateUser;
const alt = createUser;
whenElse(userExists, updateUser, createUser, user);
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** A promise that resolves to a value, that is the
result of either calling the `consequent` or the `alternative` with
`value`.

### flatmap2

The same as `flatmap`, but run two promises concurrently.

### compose

Compose two function that return promises to yield a third function that
returns a promise. The resulting composite function is denoted
`g∘f : X → Z`, defined by `(g∘f)(x) = g(f(x))` for all `x` in `X`.

**Parameters**

-   `f` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)> | [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function))** The first function to
    compose. This argument can either be a function, or a promise that resolves
    to a function. The function can either return a value or a promise for a
    value.
-   `g` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)> | [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function))** The second function to
    compose. This argument can either be a function, or a promise that resolves
    to a function. The function can either return a value or a promise for a
    value.
-   `x` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise) | any)** The argument to call `(g∘f)` with. It can either be
    any value, or the promise for any value.

**Examples**

```javascript
const f = x => future(x + 1);
const g = x => future(x + 5);
const h = compose(f, g);
h(10).then(console.log); // 10 + 1 + 5, returns 16.
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** The result of calling `g` with the result of
`f(x)`.

### flatmap3

The same as `flatmap`, but run three promises concurrently.

### unlessElse

Like `whenElse`, only call the `consequent` if the predicate returns
`false`, and the `alternative` if the predicate returns `true`;

### flatmap4

The same as `flatmap`, but run four promises concurrently.

### retry

Call an action, and retry it in case it fails. An action is retried up to
five times with an increasing timeout.

**Parameters**

-   `action` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise) \| [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function))** A function or a promise that should be
    retried in case of failure.

**Examples**

```javascript
// Retries `fetchUser` in case of failure.
retry(fetchUser).then(console.log).catch(console.error);
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** A promise that resolves to the return value of
`action`, or that is rejected with the last exception that `action` failed
with.

### flatmap5

The same as `flatmap`, but run five promises concurrently.

### when

Like `whenElse`, but have no alternative. If the predicate returns `false`,
simply return the `value`.

### unless

Like `unlessElse`, but have no alternative. If the predicate returns `true`,
simply return the `value`.

### retry2

Like `retry`, but accept one additional argument that is applied to
`action`.

**Parameters**

-   `action` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise) \| [Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function))** A function or a promise that should be
    retried in case of failure.
-   `arg` **any** The argument to apply to `action`.

**Examples**

```javascript
// Retries `fetchUser` in case of failure.
retry(fetchUser, 23).then(console.log).catch(console.error);
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;any>** A promise that resolves to the return value of
`action`, or that is rejected with the last exception that `action` failed
with.

### retry3

Like `retry2`, but accept two arguments to apply to `action`.

### map

Map a function over every element of a list. This is equivalent to
`Array.map`. Only one promise at a time get's resolved.

**Parameters**

-   `f` **[Function](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Statements/function)&lt;any>** The function that is applied to every element. This
    function can either return a value or the promise for a value.
-   `xs` **([Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)> | [Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)&lt;any>)** The list to map `f` over. This can
    either be an array, or the promise for an array.

**Examples**

```javascript
const f = x => future(x + 1);
const xs = [...Array(5).keys()];
map(f, xs).then(console.log); // Prints [1, 2, 3, 4, 5]
```

Returns **[Promise](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Promise)&lt;[Array](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array)>** A list of the same length as `xs`, but with `f`
applied to each of its elements.

### retry4

Like `retry2`, but accept three arguments to apply to `action`.

### map2

The same as `map`, but run two promises concurrently.

### map3

The same as `map`, but run three promises concurrently.

### map4

The same as `map`, but run four promises concurrently.

### map5

The same as `map`, but run five promises concurrently.
