(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(factory((global.combinatorsP = {})));
}(this, (function (exports) { 'use strict';

/**
 * Lift a value into a promise. This is equivalent to `Promise.resolve`.
 *
 * @example
 * const x = future(23);
 * const f = a => future(a + 1);
 *
 * @param {(Promise.<*>|*)} x The value to lift into a promise. This can
 * either be a value, or a promise that resolves to a value.
 * @returns {Promise.<*>} The value inside a promise.
 */
const future = x => Promise.resolve(x);

const ncurry = n => {
  const localCurry = (f, ...args) => (...largs) => {
    const rest = args.concat(largs);

    if (rest.length < n) return localCurry(f, ...rest);
    return f(...rest);
  };
  return localCurry;
};

const curry2 = ncurry(2);
const curry3 = ncurry(3);
const curry4 = ncurry(4);

const clone = (obj, hash = new WeakMap()) => {
  if (Object(obj) !== obj) return obj; // primitives
  if (hash.has(obj)) return hash.get(obj); // cyclic references
  let result;

  if (Array.isArray(obj)) result = [];
  else if (obj.constructor) result = new obj.constructor();
  else result = Object.create(null);

  hash.set(obj, result);

  if (obj instanceof Map)
    Array.from(obj, ([key, val]) => result.set(key, clone(val, hash)));
  return Object.assign(
    result,
    ...Object.keys(obj).map(key => ({[key]: clone(obj[key], hash)}))
  );
};

const promisify = f => (...args) =>
  Promise.all(args.map(future)).then(params => f(...params));

/**
 * Run a function for side effect and return the original value. The original
 * value gets cloned and therefore can be modified in the tap handler.
 *
 * @example
 * const f = a => future(a);
 * flow([f, tap(console.log)])(23); // Print "23" to the console.
 *
 * @param {(Promise.<Function>|Function)} f The function for side effect. The
 * return value of this function will be ignored.
 * @param {(Promise.<*>|*)} x The value to use for side effect. This can
 * either be a value or a promise that resolves to a value.
 * @returns {Promise.<*>} Returns `x`.
 */
const tap = curry2(promisify((f, x) => future(f(clone(x))).then(() => x)));

/**
 * Map a function over a promise.
 *
 * @param {(Promise.<Function>|Function.<*>)} f The function to apply. `f` can
 * either be a function, or a promise that resolves to a promise. The function
 * can either return a value, or a promise that resolves to a promise.
 * @param {(Promise.<*>|*)} x The value to apply to `f`. This can either be a
 * value, or a promise that resolves to a value.
 * @returns {Promise.<*>} The result of `x` applied to `f`.
 */
const fmap = curry2(promisify((f, x) => future(x).then(f)));

/**
 * Apply a function wrapped in a promise to a promisified value.
 *
 * @param {Promise.<Function>} pf A promise that resolves to a function.
 * @param {Promise.<Any>} p A promise that resolves to a value.
 * @returns {Promise.<Any>} A promise resolving to x applied to the function
 * that f resolves to.
 * @example
 * const pf = future(v => v + 1);
 * const p = future(1);
 * apply(pf, f).then(console.log); // Returns 2.
 */
const apply = curry2(promisify((pf, p) => fmap(f => fmap(f, p), pf)));

/**
 * Create a function that always returns the same value.
 *
 * @example
 * const f = constant("Hello");
 * f().then(console.log); // Returns "Hello"
 *
 * @param {(Promise.<*>|*)} x The value to return. This can either be a value
 * or a promise for a value.
 * @returns {Promise.<*>} A promise that resolves to `x`.
 */
const constant = x => () => future(x);

/**
 * Create a function that evaluates all promises in an array when called. This
 * is equivalent to `Promise.all`, with the difference that it creates a
 * callable function.
 *
 * @example
 * const f = all([openFile1(), opeFile2(), openFile3()]);
 * f().then(console.log); // Returns [a, b, c];
 *
 * @param {(Promise.<Array>|Array.<*>)} xs An array of values that are
 * resolved. `xs` can either be an array, or a promise that resolves to an
 * array. Each element of `xs` can either be a value, or a promise that
 * resolves to a value.
 * @returns {Function.<Promise.Array<*>>} A function, that when called, resolves all
 * promises and returns an Array of values.
 */
const all = xs => () => future(xs).then(ys => Promise.all(ys.map(future)));

/**
 * Compose two function that return promises to yield a third function that
 * returns a promise. The resulting composite function is denoted
 * `g∘f : X → Z`, defined by `(g∘f)(x) = g(f(x))` for all `x` in `X`.
 *
 * @example
 * const f = x => future(x + 1);
 * const g = x => future(x + 5);
 * const h = compose(f, g);
 * h(10).then(console.log); // 10 + 1 + 5, returns 16.
 *
 * @param {(Promise.<Function>|Function)} f The first function to
 * compose. This argument can either be a function, or a promise that resolves
 * to a function. The function can either return a value or a promise for a
 * value.
 * @param {(Promise.<Function>|Function)} g The second function to
 * compose. This argument can either be a function, or a promise that resolves
 * to a function. The function can either return a value or a promise for a
 * value.
 * @param {(Promise|*)} x The argument to call `(g∘f)` with. It can either be
 * any value, or the promise for any value.
 * @returns {Promise.<*>} The result of calling `g` with the result of
 * `f(x)`.
 */
const compose = curry3(promisify((f, g, x) => future(f(x)).then(g)));

/**
 * Create a function out of a list of functions, where each successive
 * invocation is supplied the return value of the previous function call. The
 * new function forms a pipe where the results flow from left to right so to
 * speak. It's a shortcut for composing more than two functions.
 *
 * @example
 * const f = (x, y) => future(x + y);
 * const fs = map(f, [...Array(5).keys()]);
 * flow(fs, 0).then(console.log); // The sum of [0, 1, 2, 3, 4], returns 10
 *
 * @param {(Promise.<Array>|Array.<Function>)} fs An array of functions to
 * compose. `fs` can either be an array, or a promise that resolves to an
 * array. Each function can either return a value or a promise for a value.
 * @param {(Promise|*)} x The argument to call the function pipeline with. It
 * can either be any value, or the promise for any value.
 * @returns {Promise.<*>} The result of applying `x` to the pipeline of
 * `fs`.
 */
const flow = curry2(
  promisify((fs, x) => fs.reduce((memo, f) => memo.then(f), future(x)))
);

/**
 * Reduce a list of values to a single value, using a reduction function. This
 * is equivalent to `Array.reduce`.
 *
 * @example
 * const f = (acc, x) => future(acc + x);
 * const xs = [...Array(5).keys()];
 * fold(f, 0, xs).then(console.log); // The sum of [0, 1, 2, 3, 4], returns 10;
 *
 * @param {(Promise.<Function>|Function.<*>)} f The reduce function to map
 * over the list. `f` can either be a function, or a promise that resolves to
 * a promise. This function can either return a value or the promise for a
 * value.
 * @param {(Promise.<*>|*)} acc The initial value to apply to the first call
 * of `f`. It can either be any value or the promise for any value.
 * @param {(Promise.<Array>|Array.<*>)} xs The list to reduce to a single
 * value. This can either be an array, or the promise for an array.
 * @returns {Promise.<*>} The value of `xs` reduced over `f`.
 */
const fold = curry3(
  promisify((f, acc, xs) =>
    xs.reduce((memo, x) => memo.then(y => f(y, x)), future(acc))
  )
);

const mapN = curry2(
  promisify(
    (concurrency, f, xs) =>
      // eslint-disable-next-line promise/avoid-new
      new Promise((resolve, reject) => {
        const ret = [];
        const iterator = xs[Symbol.iterator]();
        let isRejected = false;
        let iterableDone = false;
        let resolvingCount = 0;
        let currentIdx = 0;

        const next = () => {
          if (isRejected) return;
          const nextItem = iterator.next();
          const i = currentIdx;
          currentIdx += 1;

          if (nextItem.done) {
            iterableDone = true;

            if (resolvingCount === 0) resolve(ret);
            return;
          }
          resolvingCount += 1;

          Promise.resolve(nextItem.value)
            .then(f)
            // eslint-disable-next-line promise/always-return
            .then(val => {
              ret[i] = val;
              resolvingCount -= 1;
              next(); // eslint-disable-line promise/no-callback-in-promise
            })
            .catch(err => {
              isRejected = true;
              reject(err);
            });
        };

        for (let i = 0; i < concurrency; i += 1) {
          next();
          if (iterableDone) break;
        }
      })
  )
);

/**
 * Map a function over every element of a list. This is equivalent to
 * `Array.map`. Only one promise at a time get's resolved.
 *
 * @example
 * const f = x => future(x + 1);
 * const xs = [...Array(5).keys()];
 * map(f, xs).then(console.log); // Prints [1, 2, 3, 4, 5]
 *
 * @param {Function.<*>} f The function that is applied to every element. This
 * function can either return a value or the promise for a value.
 * @param {(Promise.<Array>|Array.<*>)} xs The list to map `f` over. This can
 * either be an array, or the promise for an array.
 * @returns {Promise.<Array>} A list of the same length as `xs`, but with `f`
 * applied to each of its elements.
 */
const map = mapN(1);

/**
 * The same as `map`, but run two promises concurrently.
 */
const map2 = mapN(2);

/**
 * The same as `map`, but run three promises concurrently.
 */
const map3 = mapN(3);

/**
 * The same as `map`, but run four promises concurrently.
 */
const map4 = mapN(4);

/**
 * The same as `map`, but run five promises concurrently.
 */
const map5 = mapN(5);

// Flattens an array of arrays into an array. It only flattens one level deep,
// which is what is required by flatmap.
var flatten = xxs => {
  const result = [];

  if (xxs == null) return result;

  xxs.forEach(xs => {
    if (Array.isArray(xs)) result.push(...xs);
    else result.push(xs);
  });

  return result;
};

const flatmapN = curry3((mapper, f, xs) => mapper(f, xs).then(flatten));

/**
 * Map a function over every element of a list and concatenate the results
 * into a single list. Only one promise at a time get's resolved.
 *
 * @example
 * const f = x => [x, x];
 * const xs = [1, 2];
 * flatmap(f, xs).then(console.log); // Prints [1, 1, 2, 2];
 *
 * @param {Function.<*>} f The function that is applied to every element. This
 * function can either return a value or the promise for a value.
 * @param {(Promise.<Array>|Array.<*>)} xs The list to map `f` over. This can
 * either be an array, or the promise for an array.
 * @returns {Promise.<Array>} The concatenation of applying every element of
 * `xs` to `f`.
 */
const flatmap = flatmapN(map);

/**
 * The same as `flatmap`, but run two promises concurrently.
 */
const flatmap2 = flatmapN(map2);

/**
 * The same as `flatmap`, but run three promises concurrently.
 */
const flatmap3 = flatmapN(map3);

/**
 * The same as `flatmap`, but run four promises concurrently.
 */
const flatmap4 = flatmapN(map4);

/**
 * The same as `flatmap`, but run five promises concurrently.
 */
const flatmap5 = flatmapN(map5);

/**
 * Catch an exception on a promise and call a exception handler. This is
 * equivalent to `Promise.catch`.
 *
 * @example
 * const f = () => new Error("Boom");
 * caught(console.error, f()); // Prints the exception.
 *
 * @param {Function} f The handler to call when the promise `p` get's
 * rejected.
 * @param {Promise.<*>|*} p A value or a promise that get's resolved.
 * @returns {Promise.<*>} A promise that resolves either to the value of `p`
 * or to the return value of `f`.
 */
const caught = curry2((f, p) => future(p).catch(f));

var identity = x => x;

/**
 * Make a conditional test and either call the left or the right branch.
 *
 * @example
 * const pred = userExists;
 * const cons = updateUser;
 * const alt = createUser;
 * whenElse(userExists, updateUser, createUser, user);
 *
 * @param {Function} predicate Test if a condition is `true` or `false`.
 * @param {Function} consequent Apply `value` to this function if the
 * predicate return `true`;
 * @param {Function} alternative Apply `value` to this function if the
 * predicate return `false`;
 * @param {Promise|*} value Test this value to decide whether to call the
 * consequent or the alternative.
 * @returns {Promise.<*>} A promise that resolves to a value, that is the
 * result of either calling the `consequent` or the `alternative` with
 * `value`.
 */
const whenElse = curry4(
  promisify((predicate, consequent, alternative, value) => {
    const chooseFn = bool =>
      bool === true ? consequent(value) : alternative(value);
    return future(predicate(value)).then(chooseFn);
  })
);

/**
 * Like `whenElse`, only call the `consequent` if the predicate returns
 * `false`, and the `alternative` if the predicate returns `true`;
 */
const unlessElse = curry4((predicate, consequent, alternative, value) =>
  whenElse(predicate, alternative, consequent, value)
);

/**
 * Like `whenElse`, but have no alternative. If the predicate returns `false`,
 * simply return the `value`.
 */
const when = curry3((predicate, consequent, value) =>
  whenElse(predicate, consequent, identity, value)
);

/**
 * Like `unlessElse`, but have no alternative. If the predicate returns `true`,
 * simply return the `value`.
 */
const unless = curry3((predicate, consequent, value) =>
  unlessElse(predicate, consequent, identity, value)
);

/**
 * Lift a binary function over two promises.
 *
 * @param {(Promise.<Function>|Function.<*, *>)} f A binary function. This
 * can either be a function, or a promise that resolves to a function. The
 * function can either return a value, or a promise that resolves to a value.
 * @param {(Promise.<*>|*)} x A value that gets lifted as the first argument
 * of `f`. This can either be a value, or a promise that resolves to a value.
 * @param {(Promise.<*>|*)} y A value that gets lifted as the first argument
 * of `f`. This can either be a value, or a promise that resolves to a value.
 * @returns {Promise.<*>} The value that f returns when applied to `x` and
 * `y`.
 * @example
 * const f = (x, y) => future(x = y);
 * const a = future(1);
 * const b = future(2);
 * lift2(f, a, b).then(console.log); // Returns 3.
 */
const lift2 = curry3(
  promisify((f, x, y) => fmap(fmap(u => v => f(u, v), x), y))
);

/**
 * Call a variadic function with the value of a promise as it's arguments. If
 * the value is an array, flatten it to the formal parameters of the
 * fulfillment handler.
 *
 * @example
 * const add = (x, y) => x + y;
 * const p = future([1, 2]);
 * spread(add, p).then(console.log); // Prints 3
 *
 * @param {(Promise.<Function>|Function.<*>)} f The function to apply to the
 * value of p. This function can either return a value or the promise for a
 * value. This function can also be a promise that resolves to a function.
 * @param {(Promise.<*>|*)} p The promise that resolves to the arguments for
 * the function. This can either be a single value, an Array or a promise
 * resolving to any of those.
 * @returns {Promise.<*>} The result of applying the value of `p` to `f`.
 */
const spread = curry2(
  promisify((f, p) =>
    future(p).then(args => {
      if (Array.isArray(args)) return f(...args);
      return f(args);
    })
  )
);

/**
 * Delay the resolution of a promise chain.
 *
 * @example
 * delay(100).then(console.log) // Waits 100 ms.
 *
 * @param {number} ms The time to wait in milliseconds.
 * @param {Promise.<*>|*} value The promise or value to resolve after the
 * delay.
 * @returns {Promise} A promise that resolved to value, or whatever value
 * resolves to.
 */
const delay = curry2((ms, value) => {
  let cancel;

  // eslint-disable-next-line promise/avoid-new
  const thunk = new Promise((resolve, reject) => {
    let timeOut = setTimeout(() => resolve(value), ms);
    cancel = () => {
      if (timeOut) {
        clearTimeout(timeOut);
        timeOut = null;
      }
      reject(new Error("Promise canceled"));
    };
  });
  thunk.cancel = cancel;
  return thunk;
});

/**
 * Test if an object is a promise.
 *
 * @example
 * const p = future(23);
 * isPromise(p); // Returns true;
 *
 * @param {*} p The object to test.
 * @returns {boolean} Returns `true` if the object is a promise, otherwise
 * `false`;
 */
const isPromise = p =>
  !!p &&
  (typeof p === "object" || typeof p === "function") &&
  typeof p.then === "function";

// times is 0-indexed. A times value of 4, actually makes 5 retries.
const retrier = (times, waitTime) => (action, ...args) => {
  const delayModifier = 1.5;
  const f = isPromise(action) ? () => action : action;

  const resolver = (iter, waitMs) => {
    if (iter <= 0) return f(...args);
    return caught(
      () =>
        delay(waitMs, null).then(() =>
          resolver(iter - 1, waitMs * delayModifier)
        ),
      f(...args)
    );
  };

  return resolver(times, waitTime);
};

/**
 * Call an action, and retry it in case it fails. An action is retried up to
 * five times with an increasing timeout.
 *
 * @example
 * // Retries `fetchUser` in case of failure.
 * retry(fetchUser).then(console.log).catch(console.error);
 *                        
 *
 * @param {Promise|Function} action A function or a promise that should be
 * retried in case of failure.
 * @returns {Promise.<*>} A promise that resolves to the return value of
 * `action`, or that is rejected with the last exception that `action` failed
 * with.
 */
const retry = retrier(4, 250);

// Using curry allows to fix the number of arguments for functions that are
// implemented as variadic functions.
/**
 * Like `retry`, but accept one additional argument that is applied to
 * `action`.
 *
 * @example
 * // Retries `fetchUser` in case of failure.
 * retry(fetchUser, 23).then(console.log).catch(console.error);
 *
 * @param {Promise|Function} action A function or a promise that should be
 * retried in case of failure.
 * @param {*} arg The argument to apply to `action`.
 * @returns {Promise.<*>} A promise that resolves to the return value of
 * `action`, or that is rejected with the last exception that `action` failed
 * with.
 */
const retry2 = curry2(retrier(4, 250));

/**
 * Like `retry2`, but accept two arguments to apply to `action`.
 */
const retry3 = curry3(retrier(4, 250));
/**
 * Like `retry2`, but accept three arguments to apply to `action`.
 */
const retry4 = curry4(retrier(4, 250));

exports.future = future;
exports.tap = tap;
exports.fmap = fmap;
exports.apply = apply;
exports.constant = constant;
exports.all = all;
exports.compose = compose;
exports.flow = flow;
exports.fold = fold;
exports.map = map;
exports.map2 = map2;
exports.map3 = map3;
exports.map4 = map4;
exports.map5 = map5;
exports.flatmap = flatmap;
exports.flatmap2 = flatmap2;
exports.flatmap3 = flatmap3;
exports.flatmap4 = flatmap4;
exports.flatmap5 = flatmap5;
exports.caught = caught;
exports.when = when;
exports.whenElse = whenElse;
exports.unless = unless;
exports.unlessElse = unlessElse;
exports.lift2 = lift2;
exports.spread = spread;
exports.delay = delay;
exports.retry = retry;
exports.retry2 = retry2;
exports.retry3 = retry3;
exports.retry4 = retry4;
exports.isPromise = isPromise;

Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=combinatorsP.min.js.map
